-- EventTimer.lua

local L = LibStub("AceLocale-3.0"):GetLocale("AshranBuddy")

local AshranBuddy = _G.AshranBuddy
if (not AshranBuddy) then
	DEFAULT_CHAT_FRAME:AddMessage(format(L["Module '%s' requires AshranBuddy"], L["Event Timer"]))
end

-- Registration
local EventTimer = AshranBuddy:RegisterModule("EventTimer")
LibStub("AceTimer-3.0"):Embed(EventTimer)

-- Locals
local Event = { }
Event[178019] = { ID = 178019, Name = L["Collect Empowered Ore"], Title = L["Empowered Ore"], MapX = math.roundup(0.545, 2), MapY = math.roundup(0.690, 2) }
Event[178021] = { ID = 178021, Name = L["Ogre Fires"], Title = L["Ogre Fires"], MapX = math.roundup(0.575, 2), MapY = math.roundup(0.307, 2) }
Event[178020] = { ID = 178020, Name = L["Risen Spirits"], Title = L["Risen Spirits"], MapX = math.roundup(0.372, 2), MapY = math.roundup(0.631, 2) }
Event[178022] = { ID = 178022, Name = L["Stadium Racing"], Title = L["Stadium Racing"], MapX = math.roundup(0.376, 2), MapY = math.roundup(0.375, 2) }

local States = { }
States.Hidden = { ID = -1, Category = 6, Value = 0, MinValue = 0, MaxValue = 0, Red = 0, Green = 0, Blue = 0, Text = "", Message = nil }
States.Inactive = { ID = 0, Category = 5, Value = 0, MinValue = 0, MaxValue = 300, Red = 0.5, Green = 0.5, Blue = 0.5, Text = nil, Message = L["%s are on cooldown for %s."] }
States.Intermission = { ID = 1, Category = 4, Value = 0, MinValue = 0, MaxValue = 360, Red = 1, Green = 0, Blue = 0, Text = nil, Message = L["%s will announce randomly within the next %s."] }
States.Announced = { ID = 2, Category = 3, Value = 0, MinValue = 0, MaxValue = 180, Red = 0.26, Green = 0.42, Blue = 1, Text = nil, Message = L["%s will begin in %s."] }
States.PreActive = { ID = 3, Category = 2, Value = 60, MinValue = 0, MaxValue = 60, Red = 0, Green = 1, Blue = 0, Text = "", Message = L["%s is currently active."] }
States.Active = { ID = 4, Category = 1, Value = 60, MinValue = 0, MaxValue = 60, Red = 0, Green = 1, Blue = 0, Text = "", Message = L["%s is currently active."] }

local Transitions = { }
Transitions[States.Inactive] = States.Intermission
Transitions[States.Intermission] = States.Hidden
Transitions[States.Announced] = States.PreActive

local LandmarkTextureIndex = 42
local IsDirty = false
local Cache = nil
local Timer = nil
local Throttle = 0

local DelayTimer = nil
local Delay = false

function EventTimer:OnInitialize()
	self:InitializeComponent(220, 65)
end

function EventTimer:OnEnable()
	self:RegisterEvent("CHAT_MSG_MONSTER_EMOTE", "OnChatMessageMonsterEmote")
	self:RegisterEvent("GROUP_JOINED", "OnGroupJoined")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "OnPlayerEnteringWorld")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnZoneChangedNewArea")
	
	AshranBuddy.db.profile.EventTimer.stacked = true
	if (not AshranBuddy.db.profile.EventTimer.locked) then
		EventTimer:Unlock()
	end
	self.component:Show()
end

function EventTimer:OnDisable()
	self:UnregisterAllEvents()
	Cache = nil
	for i = 1, #EventTimer.statusBars do
		local statusBar = self.statusBars[i]
		statusBar:SetState(States.Hidden)
	end
	if (Timer ~= nil) then
		EventTimer:CancelTimer(Timer)
		Timer = nil
	end
	self.component:Hide()
end

function EventTimer:OnChatMessageMonsterEmote(event, message, sender, language, channelString, target, flags, unused1, channelNumber, channelName, unused2, counter)
	if (AshranBuddy:IsInAshran()) then
		for i = 1, #self.statusBars do
			local statusBar = self.statusBars[i]
			if (message:find(statusBar:GetEventID())) then
				statusBar:SetState(States.Announced)
				statusBar.clean = true
				
				for j = 1, #self.statusBars do
					if (self.statusBars[j] ~= statusBar) then
						self.statusBars[j]:SetState(States.Hidden)
					end
				end
			end
		end
	end
end

function EventTimer:OnZoneChangedNewArea(event)
	if (AshranBuddy:IsInAshran()) then
		self.component:Show()
		self:SendRequest()
	else
		if (AshranBuddy.db.profile.EventTimer.locked == false) then
			self:Unlock()
		else
			self.component:Hide()
		end
		for i = 1, #EventTimer.statusBars do
			local statusBar = self.statusBars[i]
			statusBar:SetState(States.Hidden)
		end
		if (Timer ~= nil) then
			EventTimer:CancelTimer(Timer)
			Timer = nil
		end
	end
end

function EventTimer:OnPlayerEnteringWorld(event)
	if (AshranBuddy:IsInAshran()) then
		self.component:Show()
		self:SendRequest()
	else
		if (AshranBuddy.db.profile.EventTimer.locked == false) then
			self:Unlock()
		else
			self.component:Hide()
		end
		if (Timer ~= nil) then
			EventTimer:CancelTimer(Timer)
			Timer = nil
		end
	end
end

function EventTimer:OnGroupJoined(event)
	if (AshranBuddy:IsInAshran()) then
		self:SendRequest()
	end
end

function EventTimer:OnUpdate(elapsed)
	if (not AshranBuddy:IsInAshran()) then
		return
	end
	
	if (elapsed > 1.5) then
		trace("OnUpdate was suspended for", elapsed, "secs")
		for i = 1, #EventTimer.statusBars do
			local statusBar = EventTimer.statusBars[i]
			if (statusBar:GetState() ~= States.Hidden) then
				statusBar:SetValue(statusBar:GetValue() + elapsed)
				statusBar.clean = false
			end
		end
	end
	
	for i = 1, #EventTimer.statusBars do
		local statusBar = EventTimer.statusBars[i]
		local event = Event[statusBar:GetEventID()]
		local exists = EventTimer:FindWorldMapLandmark(event.Name, event.MapX, event.MapY)
		if (exists) then
			statusBar:SetState(States.Active)
			statusBar.clean = true
		elseif (not exists and statusBar:GetState() == States.Active) then
			statusBar:SetState(States.Inactive)
		end
	end
	
	if (IsDirty and AshranBuddy.db.profile.EventTimer.stacked) then
		local frame = EventTimer.component
		local bars = { }
		for i = 1, #EventTimer.statusBars do
			bars[#bars + 1] = { EventTimer.statusBars[i]:GetOrder(), EventTimer.statusBars[i] }
		end
		table.sort(bars, function(a, b) return a[1] < b[1] end)
		
		local y = -5
		for key, value in pairs(bars) do
			value[2]:ClearAllPoints()
			value[2]:SetPoint("TOPLEFT", frame, anchor, 5, y)
			y = y - 15
		end
		
		IsDirty = false
	end
	
	if (not AshranBuddy.db.profile.EventTimer.stacked) then
		local y = -5
		for i = 1, #EventTimer.statusBars do
			EventTimer.statusBars[i]:ClearAllPoints()
			EventTimer.statusBars[i]:SetPoint("TOPLEFT", EventTimer.component, "TOPLEFT", 5, y)
			y = y - 15
		end
	end
	
	Throttle = Throttle + elapsed
	if (Throttle > 1.0 and not Delay) then
		if (EventTimer:OnUpdateTimers()) then
			Delay = true
			DelayTimer = EventTimer:ScheduleTimer("OnDelayTimerCallback", 10)
		end
		Throttle = Throttle - 1.0
	end
end

function EventTimer:OnStatusBarClick(button)
	local state = self:GetState()
	local minValue, maxValue = self:GetMinMaxValues()
	local timeLeft = maxValue - self:GetValue()
	local timeText = string.format(L["%.2d:%.2d"], timeLeft / 60 % 60, timeLeft % 60)
	
	local message = ""
	if (self:GetState() == States.Inactive) then
		message = L["Events"]
	elseif (self:GetState() == States.Intermission) then
		message = L["An Event"]
	else
		message = GetSpellLink(self:GetEventID())
	end
	message = string.format(L["AshranBuddy: %s"], message)
	message = string.format(state.Message, message, timeText)

	if (button == "LeftButton" and IsShiftKeyDown() and IsInRaid()) then
		SendChatMessage(message, "RAID")
	end
	if (button == "LeftButton" and IsAltKeyDown() and IsInRaid()) then
		SendChatMessage(message, "RAID_WARNING")
	end
	if (button == "LeftButton" and IsControlKeyDown()) then
		local index = GetChannelName(L["General - Ashran"])
		if (index ~= nil) then 
			SendChatMessage(message, "CHANNEL", nil, index)
		end
	end
end

function EventTimer:OnTimerCallback(statusBar, transition)
	local value = statusBar:GetValue()
	local minValue, maxValue = statusBar:GetMinMaxValues()
	local timeLeft = maxValue - value
	
	statusBar:SetEventTime(string.format(L["%.2d:%.2d"], timeLeft / 60 % 60, timeLeft % 60))
	statusBar:SetValue(value + 1)

	if (value >= maxValue) then
		if (statusBar.eventID == 178021) then
			statusBar:SetState(States.Inactive)
		else
			statusBar:SetState(transition)
		end
	end
end

function EventTimer:OnDelayTimerCallback(statusBar, transition)
	Delay = false
end

function EventTimer:Lock()
	if (not AshranBuddy:IsInAshran()) then
		self.component:Hide()
	end
	self.handle:Hide()
	
	local point, relativeTo, relativePoint, x, y = self.component:GetPoint(1)
	AshranBuddy.db.profile.EventTimer.point = point
	AshranBuddy.db.profile.EventTimer.relativePoint = relativePoint
	AshranBuddy.db.profile.EventTimer.x = x
	AshranBuddy.db.profile.EventTimer.y = y
end

function EventTimer:Unlock()
	self.component:Show()
	self.handle:Show()
end

function EventTimer:InitializeComponent(width, height)
	self.statusBars = { }
	IsDirty = false
	
	local profile = AshranBuddy.db.profile.EventTimer
	
	local frame = self.component
	frame:SetParent(UIParent)
	frame:SetSize(width, height)
	frame:SetPoint(profile.point, UIParent, profile.relativePoint, profile.x, profile.y)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetBackdrop({ bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background" })
	frame:SetBackdropColor(0, 0, 0, 0)
	frame:SetToplevel(true)
	frame:SetClampedToScreen(true)
	
	local handle = CreateFrame("Frame", nil, frame)
	handle:SetSize(width, height)
	handle:SetPoint("TOPLEFT", frame)
	handle:Hide()

	local handleBox = handle:CreateTexture(nil, "OVERLAY", nil, -1)
	handleBox:SetTexture(0, 0, 0, 0.6)
	handleBox:SetSize(width, height)
	handleBox:SetPoint("TOPLEFT")
	
	local handleFontString = handle:CreateFontString("HandleLabel", "OVERLAY", "TextStatusBarText")
	handleFontString:SetPoint("CENTER")
	handleFontString:SetText(L["Ashran Buddy\n\nYou MUST Type /ab lock or\n/ashranbuddy lock to activate.\nBars won't show until event is announced"])
	
	-- Children
	self.handle = handle
	local offset = -5
	for key, value in pairs(Event) do
		self.statusBars[#self.statusBars + 1] = self:CreateStatusBarFrame(frame, value.ID, value.Title, value.MapX, value.MapY, offset)
		offset = offset - 15
	end
	
	-- Events
	frame:SetScript("OnUpdate", EventTimer.OnUpdate)
	frame:SetScript("OnMouseDown", function(self)
		if (AshranBuddy.db.profile.EventTimer.locked == false) then
			self:StartMoving()
		end
	end)
	frame:SetScript("OnMouseUp", function(self)
		if (AshranBuddy.db.profile.EventTimer.locked == false) then
			self:StopMovingOrSizing()
		end
	end)
end

function EventTimer:CreateStatusBarFrame(parent, id, title, mapX, mapY, offset)
	local width = parent:GetWidth() - 10

	local statusBar = CreateFrame("StatusBar", nil, parent)
	statusBar:SetWidth(width)
	statusBar:SetHeight(10)
	statusBar:SetPoint("TOPLEFT", parent, "TOPLEFT", 5, offset)
	statusBar:SetMinMaxValues(0, 180)
	statusBar:SetValue(0)
	statusBar:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar")
	statusBar:SetStatusBarColor(0.26, 0.42, 1)

	-- Artwork Layer
	local borderLeft = statusBar:CreateTexture("BorderLeft", "ARTWORK")
	borderLeft:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-Skills-BarBorder")
	borderLeft:SetTexCoord(0.007843, 0.043137, 0.193548, 0.774193)
	borderLeft:SetPoint("LEFT", -3, 0)
	borderLeft:SetSize(9, 14)

	local borderRight = statusBar:CreateTexture("BorderRight", "ARTWORK")
	borderRight:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-Skills-BarBorder")
	borderRight:SetTexCoord(0.043137, 0.007843, 0.193548, 0.774193)
	borderRight:SetPoint("RIGHT", 3, 0)
	borderRight:SetSize(9, 14)
	
	local borderMid = statusBar:CreateTexture("BorderMid", "ARTWORK")
	borderMid:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-Skills-BarBorder")
	borderMid:SetTexCoord(0.113726, 0.1490196, 0.193548, 0.774193)
	borderMid:SetPoint("TOPLEFT", borderLeft, "TOPRIGHT")
	borderMid:SetPoint("BOTTOMRIGHT", borderRight, "BOTTOMLEFT")
	borderMid:SetSize(9, 14)

	statusBar.eventNameFontString = statusBar:CreateFontString("EventNameLabel", "ARTWORK", "TextStatusBarText")
	statusBar.eventNameFontString:SetPoint("LEFT", statusBar, "LEFT", 8, 0)

	statusBar.eventTimeFontString = statusBar:CreateFontString("EventTimeLabel", "ARTWORK", "TextStatusBarText")
	statusBar.eventTimeFontString:SetPoint("RIGHT")
	
	if (not AshranBuddy.db.profile.EventTimer.showTimer) then
		eventTimeFontString:Hide()
	end
	
	-- Background Layer
	local barFull = statusBar:CreateTexture("BarFull", "BACKGROUND")
	barFull:SetVertexColor(0.04, 0.07, 0.18)
	
	statusBar.eventID = id
	statusBar.state = States.Hidden
	statusBar.mapX = mapX
	statusBar.mapY = mapY
	statusBar.clean = false
	
	statusBar.GetEventID = function(self)
		return self.eventID
	end
	statusBar.GetEventTitle = function(self)
		return self.eventNameFontString:GetText()
	end
	statusBar.SetEventTitle = function(self, value)
		self.eventNameFontString:SetText(value)
	end
	statusBar.SetEventTime = function(self, value)
		self.eventTimeFontString:SetText(value)
	end
	statusBar.GetOrder = function(self)
		return ((self.state.Category * 100000) - self:GetValue())
	end
	statusBar.GetState = function(self)
		return self.state
	end
	statusBar.SetState = function(self, state)
		self.state = state
		if (self.eventTimer ~= nil) then
			EventTimer:CancelTimer(self.eventTimer)
			self.eventTimer = nil
		end
		if (state == States.Hidden) then
			self:SetEventTime("")
			self:SetMinMaxValues(state.MinValue, state.MaxValue)
			self:SetValue(state.Value)
			self.clean = false
			self:Hide()
		elseif (AshranBuddy:IsInAshran()) then	
			if (state == States.Inactive) then
				self:SetEventTitle(L["Cooldown"])
			elseif (state == States.Intermission) then
				self:SetEventTitle(L["Be Ready!"])
			else
				self:SetEventTitle(Event[self.eventID].Title)
			end
		
			local timeLeft = state.MaxValue - state.Value
			local text = state.Text
			if (text == nil) then
				text = string.format(L["%.2d:%.2d"], timeLeft / 60 % 60, timeLeft % 60)
			end
			
			self:SetEventTime(text)
			self:SetMinMaxValues(state.MinValue, state.MaxValue)
			self:SetValue(state.Value)
			self:SetStatusBarColor(state.Red, state.Green, state.Blue)
		
			if (Transitions[state] ~= nil) then
				self.eventTimer = EventTimer:ScheduleRepeatingTimer("OnTimerCallback", 1, self, Transitions[state])
			end
			
			self:Show()
		end
		IsDirty = true
	end

	statusBar:SetScript("OnMouseDown", EventTimer.OnStatusBarClick)
	
	statusBar:SetEventTitle(title)
	statusBar:SetEventTime("")
	statusBar:Hide()
	
	-- Skin if applicable addons are loaded.
	if (IsAddOnLoaded("ElvUI") or IsAddOnLoaded("Tukui")) then
		local c
		if ElvUI then
			local E, L, V, P, G, DF = unpack(ElvUI)
			c = E
		else
			local T, C, L, G = unpack(Tukui)
			c = T
			c.TexCoords = {.08, .92, .08, .92}
		end
		statusBar:StripTextures(true)
		statusBar:CreateBackdrop()
		if c.media then
			statusBar:SetStatusBarTexture(c.media.normTex)
		elseif c.Medias then
			statusBar:SetStatusBarTexture(c.Medias.Normal)
		end
	end
	
	return statusBar
end

function EventTimer:FindWorldMapLandmark(eventName, mapX, mapY) 
	local count = GetNumMapLandmarks()
	for i = 1, count do
        local worldMapPOIName = "WorldMapFramePOI" .. i
        local worldMapPOI = _G[worldMapPOIName]
		
		local name, description, textureIndex, x, y, mapLinkID, inBattleMap, graveyardID, areaID, poiID, isObjectIcon = GetMapLandmarkInfo(i)
		if (textureIndex == LandmarkTextureIndex and mapX == math.roundup(x, 2) and mapY == math.roundup(y, 2)) then
			return true
		end
	end
	return false
end

function EventTimer:OnChatCommand(input)
	if (input == "unlock") then
		EventTimer:Unlock()
		AshranBuddy.db.profile.EventTimer.locked = false
	elseif (input == "lock") then
		EventTimer:Lock()
		AshranBuddy.db.profile.EventTimer.locked = true
	elseif (input == "sync") then
		self:SendRequest()
	elseif (input == "clear") then
		Cache = nil
		for i = 1, #EventTimer.statusBars do
			local statusBar = self.statusBars[i]
			statusBar:SetState(States.Hidden)
			statusBar:Hide()
		end
		if (Timer ~= nil) then
			EventTimer:CancelTimer(Timer)
			Timer = nil
		end
	end
end

function EventTimer:OnCreateOptions(options)
	options.EventTimer = {
		name = L["Event Timer"],
		type = "group",
		order = 10,
		args = {
			header = {
				name = L["Event Timer"],
				type = "header",
				order = 0
			},
			headerDescription = {
				name = L["Tracks and synchronizes timers for all Ashran events."],
				type = "description",
				order = 1
			},
			enabled = {
				name = L["Enabled"],
				desc = L["Enables the Ashran Event Timer module."],
				type = "toggle",
				set = function(info, val) AshranBuddy.db.profile.EventTimer.enabled = val; if (val) then self:Enable(); self:SendRequest() else self:Disable() end end,
				get = function(info) return AshranBuddy.db.profile.EventTimer.enabled end,
				order = 2
			},
			enabledDescription = {
				name = L["Enables the Ashran Event Timer module."],
				type = "description",
				order = 3
			},
			locked = {
				name = L["Lock Event Timer"],
				desc = L["Locks and unlocks the event timer frame."],
				type = "toggle",
				set = function(info, val) AshranBuddy.db.profile.EventTimer.locked = val; if (AshranBuddy.db.profile.EventTimer.locked) then self:Lock() else self:Unlock() end end,
				get = function(info) return AshranBuddy.db.profile.EventTimer.locked end,
				order = 4
			},
			lockedDescription = {
				name = L["Locks and unlocks the event timer frame."],
				type = "description",
				order = 5
			},
			showTimer = {
				name = L["Show Numbered Timer"],
				desc = L["Displays numbered timer on the right of the countdown bar."],
				type = "toggle",
				set = function(info, val) AshranBuddy.db.profile.EventTimer.showTimer = val; if (AshranBuddy.db.profile.EventTimer.showTimer) then for i = 1, #self.statusBars do self.statusBars[i].eventTimeFontString:Show() end else for i = 1, #self.statusBars do self.statusBars[i].eventTimeFontString:Hide() end end end,
				get = function(info) return AshranBuddy.db.profile.EventTimer.showTimer end,
				order = 6
			},
			showTimerDescription = {
				name = L["Displays numbered timer on the right of the countdown bar."],
				type = "description",
				order = 7
			},
		},
	}
end

function EventTimer:OnMessageReceived(packet, distribution, sender)
	-- A raid member has requested a synchronise
	if (packet.type == "request") then
		EventTimer:SendResponse(sender)
	end
	
	-- You requested a synchronise and here it is (from cross realm player)
	local name, realm = UnitName("player")
	if (packet.type == "synchronise" and distribution == "RAID" and packet.caller:find(name)) then
		EventTimer:ReceiveSynchronise(sender, packet.data)
	end
	
	-- You requested a synchronise and here it is
	if (packet.type == "synchronise" and distribution == "WHISPER") then
		EventTimer:ReceiveSynchronise(sender, packet.data)
	end
end

function EventTimer:SendRequest()
	if (AshranBuddy:IsInAshran() and IsInRaid()) then
		local packet = { type = "request", data = {} }
		self:SendMessage(packet, "RAID")
	end
end

function EventTimer:SendResponse(sender)
	if (AshranBuddy:IsInAshran() and EventTimer.enabled) then
		local packet = { type = "synchronise", data = { } }
		for i = 1, #EventTimer.statusBars do
			local statusBar = EventTimer.statusBars[i]
			packet.data[i] = { state = statusBar:GetState().ID, value = statusBar:GetValue(), vr = AshranBuddy.version }
		end
		local name, realm = UnitName(sender, true)
		if (realm == nil or realm == "") then
			-- Player is on the same server
			trace("Sending data to:", sender, "(Same Realm)")
			EventTimer:SendMessage(packet, "WHISPER", sender)
		else
			-- Player is cross realm
			trace("Sending data to:", sender, "(Cross Realm)")
			packet.caller = sender
			EventTimer:SendMessage(packet, "RAID")
		end
	end
end

function EventTimer:ReceiveSynchronise(sender, data)
	if (AshranBuddy:IsInAshran() and data ~= nil and sender ~= UnitName("player") and not Delay) then
		trace("Receiving data from:", UnitName(sender))
		if (Cache == nil) then
			Cache = { }
		end
		
		local states = { }
		for key, value in pairs(States) do
			states[value.ID] = value
		end
		
		local count = 0
		for i = 1, #data do
			data[i].state = states[data[i].state]
			trace("Receive:", sender, self.statusBars[i]:GetEventTitle(), data[i].value)
			if (data[i].state ~= States.Hidden) then
				data[i].index = i
				data[i].tolerance = math.nearest(data[i].value + (data[i].state.Category * 10000), 5) -- +/- 2secs
				data[i].sender = sender
				if (data[i].vr == nil) then
					data[i].vr = 1
				end
				table.insert(Cache, data[i])
				count = count + 1
			end
		end
		
		if (count > 0) then
			trace(count, "items added to cache from", sender)
		end
	end
end

function EventTimer:OnUpdateTimers()
	local complete = false
	if (AshranBuddy:IsInAshran() and Cache ~= nil) then
		trace("Start updating timers...")
		for i = 1, #EventTimer.statusBars do
			local statusBar = EventTimer.statusBars[i]
			if (not statusBar.clean) then
				local data = table.select(Cache, function(k, a) if (a.index == i) then return a end end)
				table.sort(data, function(a, b) return a.tolerance > b.tolerance end)
				trace(#data, "items found in cache for", statusBar:GetEventTitle())
				if (#data > 0) then
					local tolerance = 0
					local occurrences = { }
					for j = 1, #data do
						local item = table.first(occurrences, function(a) return a.tolerance == data[j].tolerance end)
						if (item == nil) then
							occurrences[#occurrences + 1] = { tolerance = data[j].tolerance, count = 1, hash = data[j].tolerance }
						else
							item.count = item.count + 1
							item.hash = item.tolerance * item.count
						end
					end
					table.sort(occurrences, function(a, b) return a.hash > b.hash end)
					
					local count = 0
					for j = 1, #occurrences do
						trace("Occurrences:", occurrences[j].tolerance, "#"..tostring(occurrences[j].count))
						if (occurrences[j].count > count) then
							tolerance = occurrences[j].tolerance
							count = occurrences[j].count
						end
					end
					
					if (tolerance ~= 0) then
						trace("Tolerance: ", tolerance)
						local data = array.select(data, function(a) return a.tolerance == tolerance end)
						table.sort(data, function(a, b) return a.vr > b.vr end)
						for j = 1, #data do
							trace("Available:", data[j].index, data[j].tolerance, "v"..tostring(data[j].vr), data[j].sender)
						end
						
						local status = array.first(data, function(a) return a.tolerance == tolerance end)
						if (status ~= nil) then
							trace("Used:", status.index, status.tolerance, status.vr, status.sender, "|", status.value)
							
							statusBar:SetState(status.state)
							statusBar:SetValue(status.value)
						else
							trace("Used: None, no matching tolerance found")
						end
					else
						trace("Tolerance: 0, no valid occurrences")
					end
				end
			end
		end
		trace("End updating timers...")
		complete = true
	end
	Cache = nil
	return complete
end